# SPDX-FileCopyrightText: 2024-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import os
from tempfile import NamedTemporaryFile
from xml.etree import ElementTree

import pytest
from conftest import (
    create_project,
)

from idf_build_apps import App
from idf_build_apps.args import (
    BuildArguments,
    DependencyDrivenBuildArguments,
    FindArguments,
    FindBuildArguments,
    expand_vars,
)
from idf_build_apps.constants import IDF_BUILD_APPS_TOML_FN, PREVIEW_TARGETS, SUPPORTED_TARGETS
from idf_build_apps.main import main
from idf_build_apps.manifest.manifest import DEFAULT_BUILD_TARGETS, FolderRule, reset_default_build_targets


def test_init_attr_deprecated_by():
    args = DependencyDrivenBuildArguments(
        ignore_app_dependencies_filepatterns=['bar'],
        modified_files=['barfile'],
    )
    assert args.deactivate_dependency_driven_build_by_filepatterns == ['bar']

    args = DependencyDrivenBuildArguments(
        deactivate_dependency_driven_build_by_components=['foo'],
        modified_components=['foocomp'],
    )
    assert args.deactivate_dependency_driven_build_by_components == ['foo']


@pytest.mark.parametrize(
    'kwargs, expected',
    [
        ({'config': 'foo'}, ['foo']),
        ({'config_rules_str': 'bar'}, ['bar']),
        ({'config_rules': ['baz']}, ['baz']),
    ],
)
def test_init_attr_override(kwargs, expected):
    args = FindBuildArguments(
        **kwargs,
    )
    assert args.config_rules == expected


def test_apply_config_with_deprecated_names():
    with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
        fw.write("""config = [
    "foo"
]
""")

    args = FindBuildArguments()
    assert args.config_rules == ['foo']


def test_apply_config_in_parent_dir(tmp_path):
    test_under = tmp_path / 'test_under'
    test_under.mkdir()
    os.chdir(test_under)

    with open(tmp_path / IDF_BUILD_APPS_TOML_FN, 'w') as fw:
        fw.write('target = "esp32"')

    assert FindArguments().target == 'esp32'


def test_apply_config_over_pyproject_toml(tmp_path):
    test_under = tmp_path / 'test_under'
    test_under.mkdir()
    os.chdir(test_under)

    with open(test_under / 'pyproject.toml', 'w') as fw:
        fw.write("""[tool.idf-build-apps]
target = "esp32s2"
""")

    assert FindArguments().target == 'esp32s2'

    with open(tmp_path / IDF_BUILD_APPS_TOML_FN, 'w') as fw:
        fw.write('target = "esp32"')

    assert FindArguments().target == 'esp32'


def test_empty_argument():
    args = FindArguments()
    assert args.config_rules is None


def test_build_args_expansion(monkeypatch):
    monkeypatch.setenv('FOO', '2')

    args = BuildArguments(
        parallel_index=2,
        parallel_count='$FOO',
        collect_app_info='@p.txt',
        junitxml='x_@p.txt',
        collect_size_info='@p_@p.txt',
    )
    assert args.collect_app_info == '2.txt'
    assert args.junitxml == 'x_2.txt'
    assert args.parallel_count == 2

    args.parallel_index = 3
    assert args.collect_app_info == '3.txt'
    assert args.junitxml == 'x_3.txt'
    assert args.collect_size_info == '3_3.txt'


def test_func_overwrite_config():
    with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
        fw.write("""config = [
        "foo"
    ]
    modified_components = [
        'comp1',
    ]
    no_color = true
    """)

    args = FindArguments(
        config=['bar'],
    )

    assert args.config_rules == ['bar']
    assert args.modified_components == ['comp1']
    assert args.modified_files is None


def test_func_overwrite_toml_overwrite_pyproject_toml():
    with open('pyproject.toml', 'w') as fw:
        fw.write("""[tool.idf-build-apps]
config = [
    "foo"
]
modified_components = [
    'comp1',
]
ignore_app_dependencies_components = [
    'baz'
]
verbose=3
    """)

    with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
        fw.write("""config = [
    "bar"
]
modified_files = [
    'file1',
]
        """)

    args = FindArguments(
        modified_components=['comp2'],
    )
    assert args.config_rules == ['bar']
    assert args.modified_components == ['comp2']
    assert args.modified_files == ['file1']
    assert args.verbose == 3
    assert args.deactivate_dependency_driven_build_by_components == ['baz']


def test_combination_validation():
    """Test that target options can now be combined"""
    # Mock targets for consistent testing
    mock_supported = ['esp32', 'esp32s2']
    mock_preview = ['esp32h2']
    mock_all = mock_supported + mock_preview

    import idf_build_apps.args
    import idf_build_apps.constants

    original_supported = idf_build_apps.constants.SUPPORTED_TARGETS
    original_preview = idf_build_apps.constants.PREVIEW_TARGETS
    original_all = idf_build_apps.constants.ALL_TARGETS

    try:
        idf_build_apps.constants.SUPPORTED_TARGETS = mock_supported
        idf_build_apps.constants.PREVIEW_TARGETS = mock_preview
        idf_build_apps.constants.ALL_TARGETS = mock_all
        idf_build_apps.args.SUPPORTED_TARGETS = mock_supported
        idf_build_apps.args.PREVIEW_TARGETS = mock_preview
        idf_build_apps.args.ALL_TARGETS = mock_all

        # Test combination: default + preview
        args = FindBuildArguments(enable_preview_targets=True, default_build_targets=['esp32'], paths=['.'])

        # Should combine: ['esp32'] + ['esp32h2'] = ['esp32', 'esp32h2']
        expected = ['esp32', 'esp32h2']
        assert args.default_build_targets == expected

    finally:
        idf_build_apps.constants.SUPPORTED_TARGETS = original_supported
        idf_build_apps.constants.PREVIEW_TARGETS = original_preview
        idf_build_apps.constants.ALL_TARGETS = original_all
        idf_build_apps.args.SUPPORTED_TARGETS = original_supported
        idf_build_apps.args.PREVIEW_TARGETS = original_preview
        idf_build_apps.args.ALL_TARGETS = original_all


def test_build_targets_cli(tmp_path, monkeypatch):
    create_project('foo', tmp_path)
    with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
        fw.write("""paths = ["foo"]
build_dir = "build_@t"
junitxml = "test.xml"
dry_run = true
""")

    def get_enabled_targets(args):
        with monkeypatch.context() as m:
            m.setattr('sys.argv', args)
            main()
        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        testcases = test_suite.findall('testcase')
        # get targets
        return [c.attrib['name'].replace('foo/build_', '') for c in testcases]

    # default build SUPPORTED_TARGETS
    targets = get_enabled_targets(['idf-build-apps', 'build'])
    assert len(targets) == len(SUPPORTED_TARGETS)
    assert set(targets) == set(SUPPORTED_TARGETS)
    # build with --target
    targets = get_enabled_targets(['idf-build-apps', 'build', '--target', 'esp32'])
    assert len(targets) == 1
    assert targets[0] == 'esp32'
    # build with --default-build-targets
    targets = get_enabled_targets(['idf-build-apps', 'build', '--default-build-targets', 'esp32', 'esp32s2'])
    assert len(targets) == 2
    assert set(targets) == {'esp32', 'esp32s2'}
    # build with --enable-preview-targets
    targets = get_enabled_targets(['idf-build-apps', 'build', '--enable-preview-targets'])
    assert len(targets) == len(SUPPORTED_TARGETS) + len(PREVIEW_TARGETS)
    assert set(targets) == set(SUPPORTED_TARGETS) | set(PREVIEW_TARGETS)
    # build with --disable-targets
    assert 'esp32' in SUPPORTED_TARGETS and 'esp32s2' in SUPPORTED_TARGETS
    targets = get_enabled_targets(['idf-build-apps', 'build', '--disable-targets', 'esp32', 'esp32s2'])
    assert set(targets) == set(SUPPORTED_TARGETS) - {'esp32', 'esp32s2'}
    # build with --enable-preview-targets and --disable-targets
    targets = get_enabled_targets(
        ['idf-build-apps', 'build', '--enable-preview-targets', '--disable-targets', PREVIEW_TARGETS[0]]
    )
    assert len(targets) == len(SUPPORTED_TARGETS) + len(PREVIEW_TARGETS) - 1
    assert set(targets) == set(SUPPORTED_TARGETS) | set(PREVIEW_TARGETS) - {PREVIEW_TARGETS[0]}


class TestIgnoreWarningFile:
    def test_deprecated_cli(self, monkeypatch, capsys):
        with open('foo.txt', 'w') as fw:
            fw.write('warning:xxx')

        with monkeypatch.context() as m:
            m.setattr('sys.argv', ['idf-build-apps', 'build', '--ignore-warning-file', 'foo.txt'])
            main()

        assert len(App.IGNORE_WARNS_REGEXES) == 1
        assert App.IGNORE_WARNS_REGEXES[0].pattern == 'warning:xxx'

        with open('bar.txt', 'w') as fw:
            fw.write('warning:yyy')

        with monkeypatch.context() as m:
            m.setattr('sys.argv', ['idf-build-apps', 'build', '--ignore-warning-file', 'foo.txt', 'bar.txt'])
            with pytest.raises(SystemExit):
                main()

        assert 'unrecognized arguments: bar.txt' in capsys.readouterr().err

    def test_new_cli(self, monkeypatch):
        with open('foo.txt', 'w') as fw:
            fw.write('warning:xxx')
        with open('bar.txt', 'w') as fw:
            fw.write('warning:yyy')

        with monkeypatch.context() as m:
            m.setattr('sys.argv', ['idf-build-apps', 'build', '--ignore-warning-files', 'foo.txt', 'bar.txt'])
            main()

        assert len(App.IGNORE_WARNS_REGEXES) == 2
        assert App.IGNORE_WARNS_REGEXES[0].pattern == 'warning:xxx'
        assert App.IGNORE_WARNS_REGEXES[1].pattern == 'warning:yyy'

    def test_func_with_str(self):
        with open('foo.txt', 'w') as fw:
            fw.write('warning:xxx')
        with open('bar.txt', 'w') as fw:
            fw.write('warning:yyy')

        BuildArguments(
            ignore_warning_files=['foo.txt', 'bar.txt'],
        )

        assert len(App.IGNORE_WARNS_REGEXES) == 2
        assert App.IGNORE_WARNS_REGEXES[0].pattern == 'warning:xxx'
        assert App.IGNORE_WARNS_REGEXES[1].pattern == 'warning:yyy'

    def test_func_with_fs(self):
        with open('foo.txt', 'w') as fw:
            fw.write('warning:xxx')
        with open('bar.txt', 'w') as fw:
            fw.write('warning:yyy')

        BuildArguments(
            ignore_warning_file=[open('foo.txt'), open('bar.txt')],
        )

        assert len(App.IGNORE_WARNS_REGEXES) == 2
        assert App.IGNORE_WARNS_REGEXES[0].pattern == 'warning:xxx'
        assert App.IGNORE_WARNS_REGEXES[1].pattern == 'warning:yyy'

    def test_ignore_extra_fields(self):
        with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
            fw.write("""dry_run = true""")

        args = FindArguments()
        assert not hasattr(args, 'dry_run')

    def test_config_file(self, tmp_path, monkeypatch):
        create_project('foo', tmp_path)

        with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
            fw.write("""paths = ["foo"]
target = "esp32"
build_dir = "build_@t"
junitxml = "test.xml"
keep_going = true
""")

        # test basic config
        with monkeypatch.context() as m:
            m.setenv('PATH', 'foo')  # let build fail
            m.setattr('sys.argv', ['idf-build-apps', 'build'])
            with pytest.raises(SystemExit):
                main()

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['failures'] == '1'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '0'
        assert test_suite.findall('testcase')[0].attrib['name'] == 'foo/build_esp32'

        # test cli overrides config
        with monkeypatch.context() as m:
            m.setenv('PATH', 'foo')  # let build fail
            m.setattr('sys.argv', ['idf-build-apps', 'build', '--build-dir', 'build_hi_@t'])
            with pytest.raises(SystemExit):
                main()

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['failures'] == '1'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '0'
        assert test_suite.findall('testcase')[0].attrib['name'] == 'foo/build_hi_esp32'

        # test cli action_true
        with monkeypatch.context() as m:
            m.setattr('sys.argv', ['idf-build-apps', 'build', '--dry-run'])
            main()

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '1'
        assert test_suite.findall('testcase')[0].attrib['name'] == 'foo/build_esp32'

        # test config store_true set to true
        with open(IDF_BUILD_APPS_TOML_FN, 'a') as fw:
            fw.write('\ndry_run = true\n')

        with monkeypatch.context() as m:
            m.setattr('sys.argv', ['idf-build-apps', 'build'])
            main()

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '1'
        assert test_suite.findall('testcase')[0].attrib['name'] == 'foo/build_esp32'

        # test config store_true set to false, but CLI set to true
        with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
            fw.write("""paths = ["foo"]
build_dir = "build_@t"
junitxml = "test.xml"
dry_run = false
""")

        with monkeypatch.context() as m:
            m.setattr(
                'sys.argv', ['idf-build-apps', 'build', '--default-build-targets', 'esp32', 'esp32s2', '--dry-run']
            )
            main()

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '2'
        assert test_suite.findall('testcase')[0].attrib['name'] == 'foo/build_esp32'
        assert test_suite.findall('testcase')[1].attrib['name'] == 'foo/build_esp32s2'

    def test_config_file_by_cli(self, tmp_path, monkeypatch):
        create_project('foo', tmp_path)
        create_project('bar', tmp_path)

        with open(IDF_BUILD_APPS_TOML_FN, 'w') as fw:
            fw.write('paths = ["foo"]')

        with NamedTemporaryFile(mode='w', suffix='.toml') as ft:
            ft.write('paths = ["bar"]')
            ft.flush()

            with monkeypatch.context() as m:
                m.setattr(
                    'sys.argv',
                    [
                        'idf-build-apps',
                        'build',
                        '-t',
                        'esp32',
                        '--config-file',
                        ft.name,
                        '--junitxml',
                        'test.xml',
                        '--dry-run',
                    ],
                )
                main()

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())
        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '1'
        assert test_suite.findall('testcase')[0].attrib['name'] == 'bar/build'


class TestDefaultBuildTargetsContextVar:
    SUPPORTED_TARGETS = ['esp32', 'esp32s2']
    PREVIEW_TARGETS = ['esp32h2', 'esp32p4']
    ALL_TARGETS = SUPPORTED_TARGETS + PREVIEW_TARGETS + ['esp32s3', 'esp32c3']

    @pytest.fixture(autouse=True)
    def setup(self, monkeypatch):
        import idf_build_apps.args
        import idf_build_apps.constants
        import idf_build_apps.manifest.manifest

        # reset_default_build_targets() need .manifest.manifest
        monkeypatch.setattr(idf_build_apps.manifest.manifest, 'SUPPORTED_TARGETS', self.SUPPORTED_TARGETS)
        monkeypatch.setattr(idf_build_apps.constants, 'SUPPORTED_TARGETS', self.SUPPORTED_TARGETS)
        monkeypatch.setattr(idf_build_apps.args, 'SUPPORTED_TARGETS', self.SUPPORTED_TARGETS)

        monkeypatch.setattr(idf_build_apps.constants, 'PREVIEW_TARGETS', self.PREVIEW_TARGETS)
        monkeypatch.setattr(idf_build_apps.args, 'PREVIEW_TARGETS', self.PREVIEW_TARGETS)

        monkeypatch.setattr(idf_build_apps.manifest.manifest, 'ALL_TARGETS', self.SUPPORTED_TARGETS)
        monkeypatch.setattr(idf_build_apps.constants, 'ALL_TARGETS', self.ALL_TARGETS)
        monkeypatch.setattr(idf_build_apps.args, 'ALL_TARGETS', self.ALL_TARGETS)

        reset_default_build_targets()

        yield

    def test_direct_contextvar_access(self):
        assert DEFAULT_BUILD_TARGETS.get() == self.SUPPORTED_TARGETS

        # Test setting new values
        test_targets = ['esp32', 'esp32s2']
        DEFAULT_BUILD_TARGETS.set(test_targets)
        assert DEFAULT_BUILD_TARGETS.get() == test_targets

        # Test setting to ALL_TARGETS
        DEFAULT_BUILD_TARGETS.set(self.ALL_TARGETS)
        assert DEFAULT_BUILD_TARGETS.get() == self.ALL_TARGETS

    def test_folder_rule_backward_compatibility(self):
        # Test setting via contextvar
        other_targets = ['esp32h2', 'esp32p4']
        DEFAULT_BUILD_TARGETS.set(other_targets)
        assert FolderRule.DEFAULT_BUILD_TARGETS == other_targets
        assert DEFAULT_BUILD_TARGETS.get() == other_targets

        # Test setting via FolderRule
        test_targets = ['esp32c3', 'esp32c6']
        FolderRule.DEFAULT_BUILD_TARGETS = test_targets
        assert DEFAULT_BUILD_TARGETS.get() == test_targets
        assert FolderRule.DEFAULT_BUILD_TARGETS == test_targets

    def test_default_behavior(self):
        args = FindBuildArguments(paths=['.'])

        assert args.enable_preview_targets is False
        assert args.default_build_targets == self.SUPPORTED_TARGETS
        assert DEFAULT_BUILD_TARGETS.get() == self.SUPPORTED_TARGETS

    def test_default_build_targets_option(self):
        test_targets = ['esp32', 'esp32c3', 'esp32s2']
        args = FindBuildArguments(default_build_targets=test_targets, paths=['.'])

        assert args.default_build_targets == test_targets
        assert DEFAULT_BUILD_TARGETS.get() == test_targets

    def test_enable_preview_targets_option(self):
        args = FindBuildArguments(enable_preview_targets=True, paths=['.'])

        assert args.enable_preview_targets is True
        expected_targets = sorted(self.SUPPORTED_TARGETS + self.PREVIEW_TARGETS)
        assert args.default_build_targets == expected_targets
        assert DEFAULT_BUILD_TARGETS.get() == expected_targets

    def test_disable_targets_with_default_build_targets(self):
        args = FindBuildArguments(
            default_build_targets=['esp32', 'esp32s2', 'esp32c3'], disable_targets=['esp32s2'], paths=['.']
        )

        expected_targets = ['esp32', 'esp32c3']
        assert args.default_build_targets == expected_targets
        assert DEFAULT_BUILD_TARGETS.get() == expected_targets

    def test_disable_targets_with_enable_preview_targets(self):
        disabled_target = 'esp32h2'
        args = FindBuildArguments(enable_preview_targets=True, disable_targets=[disabled_target], paths=['.'])

        expected_targets = sorted(
            [_t for _t in [*self.SUPPORTED_TARGETS, *self.PREVIEW_TARGETS] if _t != disabled_target]
        )
        assert args.default_build_targets == expected_targets
        assert DEFAULT_BUILD_TARGETS.get() == expected_targets

    def test_invalid_targets_filtering(self):
        invalid_targets = ['esp32', 'invalid_target', 'esp32s2', 'another_invalid']
        args = FindBuildArguments(default_build_targets=invalid_targets, paths=['.'])

        # Only valid targets should remain
        expected_targets = ['esp32', 'esp32s2']
        assert args.default_build_targets == expected_targets
        assert DEFAULT_BUILD_TARGETS.get() == expected_targets

    def test_contextvar_isolation_between_instances(self):
        """Test that the contextvar behaves correctly across multiple argument instances"""
        # First instance sets default_build_targets
        FindBuildArguments(default_build_targets=['esp32', 'esp32s2'], paths=['.'])
        assert DEFAULT_BUILD_TARGETS.get() == ['esp32', 'esp32s2']

        # Second instance sets enable_preview_targets (should combine with SUPPORTED_TARGETS)
        FindBuildArguments(enable_preview_targets=True, paths=['.'])
        assert DEFAULT_BUILD_TARGETS.get() == sorted(self.SUPPORTED_TARGETS + self.PREVIEW_TARGETS)

        FindBuildArguments(paths=['.'])
        assert DEFAULT_BUILD_TARGETS.get() == self.SUPPORTED_TARGETS

    def test_empty_default_build_targets(self):
        args = FindBuildArguments(default_build_targets=[])

        # Empty list is treated as falsy, so it falls back to default behavior
        assert args.default_build_targets == self.SUPPORTED_TARGETS
        assert DEFAULT_BUILD_TARGETS.get() == self.SUPPORTED_TARGETS

    def test_additional_build_targets_option(self):
        """Test that --additional-build-targets option works correctly"""
        additional_targets = ['esp32h2', 'esp32p4']
        args = FindBuildArguments(additional_build_targets=additional_targets, paths=['.'])

        expected_targets = sorted(self.SUPPORTED_TARGETS + additional_targets)
        assert args.additional_build_targets == additional_targets
        assert args.default_build_targets == expected_targets
        assert DEFAULT_BUILD_TARGETS.get() == expected_targets

    def test_additional_build_targets_with_existing_target(self):
        existing_target = 'esp32'  # From self.SUPPORTED_TARGETS
        new_target = 'esp32h2'  # From self.ALL_TARGETS but not in self.SUPPORTED_TARGETS
        additional_targets = [existing_target, new_target]

        args = FindBuildArguments(additional_build_targets=additional_targets, paths=['.'])

        # Should only add the new target, not duplicate the existing one
        expected_targets = sorted([*self.SUPPORTED_TARGETS, new_target])
        assert args.additional_build_targets == additional_targets
        assert args.default_build_targets == expected_targets
        assert DEFAULT_BUILD_TARGETS.get() == expected_targets

    def test_additional_build_targets_with_invalid_target(self):
        invalid_targets = [
            'esp32',
            'invalid_target',
            'esp32h2',
        ]  # esp32 is already supported, invalid_target doesn't exist

        args = FindBuildArguments(additional_build_targets=invalid_targets, paths=['.'])

        # Should only add valid targets not already in SUPPORTED_TARGETS
        expected_targets = sorted([*self.SUPPORTED_TARGETS, 'esp32h2'])
        assert args.additional_build_targets == invalid_targets
        assert args.default_build_targets == expected_targets
        assert DEFAULT_BUILD_TARGETS.get() == expected_targets

    def test_disable_targets_with_additional_build_targets(self):
        additional_target = 'esp32h2'  # From self.ALL_TARGETS but not in self.SUPPORTED_TARGETS
        disabled_target = 'esp32'  # From self.SUPPORTED_TARGETS

        args = FindBuildArguments(
            additional_build_targets=[additional_target], disable_targets=[disabled_target], paths=['.']
        )

        expected_targets = ['esp32h2', 'esp32s2']
        assert args.default_build_targets == expected_targets
        assert DEFAULT_BUILD_TARGETS.get() == expected_targets

    def test_combination_default_and_additional(self):
        args = FindBuildArguments(
            default_build_targets=['esp32', 'esp32s2'], additional_build_targets=['esp32h2'], paths=['.']
        )

        expected_targets = ['esp32', 'esp32h2', 'esp32s2']
        assert args.default_build_targets == expected_targets
        assert DEFAULT_BUILD_TARGETS.get() == expected_targets

    def test_combination_preview_and_additional(self):
        args = FindBuildArguments(enable_preview_targets=True, additional_build_targets=['esp32p4'], paths=['.'])

        # Should combine: SUPPORTED_TARGETS + PREVIEW_TARGETS + additional
        # self.PREVIEW_TARGETS contains 'esp32p4', so it's not really additional
        expected_targets = sorted(self.SUPPORTED_TARGETS + self.PREVIEW_TARGETS)
        assert args.default_build_targets == expected_targets
        assert DEFAULT_BUILD_TARGETS.get() == expected_targets

    def test_combination_all_three_options(self):
        args = FindBuildArguments(
            default_build_targets=['esp32'],
            enable_preview_targets=True,
            additional_build_targets=['esp32p4', 'esp32c3'],
            paths=['.'],
        )

        # Should combine: ['esp32'] + PREVIEW_TARGETS + ['esp32p4', 'esp32c3']
        expected_targets = sorted({'esp32', *self.PREVIEW_TARGETS, 'esp32p4', 'esp32c3'})
        assert args.default_build_targets == expected_targets
        assert DEFAULT_BUILD_TARGETS.get() == expected_targets

    def test_empty_additional_build_targets(self):
        args = FindBuildArguments(additional_build_targets=[], paths=['.'])

        assert args.additional_build_targets == []
        assert args.default_build_targets == self.SUPPORTED_TARGETS
        assert DEFAULT_BUILD_TARGETS.get() == self.SUPPORTED_TARGETS


def test_expand_vars(monkeypatch):
    assert expand_vars('Value is $TEST_VAR') == 'Value is '
    monkeypatch.setenv('TEST_VAR', 'test_value')
    assert expand_vars('Value is $TEST_VAR') == 'Value is test_value'
    assert expand_vars('Value is $TEST_VAR and $NON_EXISTING_VAR') == 'Value is test_value and '
    assert expand_vars('No variables here') == 'No variables here'
    assert expand_vars('') == ''


class TestManifestFilepatterns:
    def test_manifest_filepatterns_default(self, tmp_path):
        root_dir = tmp_path / 'project'

        (root_dir / 'app1' / 'manifest.yml').parent.mkdir(parents=True, exist_ok=True)
        (root_dir / 'app1' / 'manifest.yml').write_text('app1:\n  enable:\n    - if: IDF_TARGET == "esp32"')

        (root_dir / 'tests' / 'manifest.yml').parent.mkdir(parents=True, exist_ok=True)
        (root_dir / 'tests' / 'manifest.yml').write_text('tests:\n  enable:\n    - if: IDF_TARGET == "esp32"')

        (root_dir / 'build' / 'managed_components' / 'comp' / 'manifest.yml').parent.mkdir(parents=True, exist_ok=True)
        (root_dir / 'build' / 'managed_components' / 'comp' / 'manifest.yml').write_text(
            'build:\n  enable:\n    - if: IDF_TARGET == "esp32"'
        )

        # Change to the root directory
        os.chdir(root_dir)

        # Test with manifest_filepatterns and default exclude patterns
        args = DependencyDrivenBuildArguments(manifest_filepatterns=['**/*.yml'], manifest_rootpath=str(root_dir))

        # Should include app1 and app2 manifest files, but exclude managed_components
        manifest_files = args.manifest_files
        assert manifest_files is not None
        assert len(manifest_files) == 2

        # Convert to relative paths for easier assertion
        rel_paths = [os.path.relpath(f, str(root_dir)) for f in manifest_files]
        assert 'app1/manifest.yml' in rel_paths
        assert 'tests/manifest.yml' in rel_paths

    def test_manifest_exclude_regexes_custom(self, tmp_path):
        """Test custom exclude patterns"""
        root_dir = tmp_path / 'project'
        root_dir.mkdir()

        # Create various manifest files
        (root_dir / 'app1' / 'manifest.yml').parent.mkdir()
        (root_dir / 'app1' / 'manifest.yml').write_text('app1:\n  enable:\n    - if: IDF_TARGET == "esp32"')

        (root_dir / 'tests' / 'manifest.yml').parent.mkdir()
        (root_dir / 'tests' / 'manifest.yml').write_text('tests:\n  enable:\n    - if: IDF_TARGET == "esp32"')

        (root_dir / 'build' / 'manifest.yml').parent.mkdir()
        (root_dir / 'build' / 'manifest.yml').write_text('build:\n  enable:\n    - if: IDF_TARGET == "esp32"')

        os.chdir(root_dir)

        # Test with custom exclude patterns
        args = DependencyDrivenBuildArguments(
            manifest_filepatterns=['**/*.yml'],
            manifest_exclude_regexes=['/tests/', '/build/'],
            manifest_rootpath=str(root_dir),
        )

        manifest_files = args.manifest_files
        assert manifest_files is not None
        assert len(manifest_files) == 1

        rel_paths = [os.path.relpath(f, str(root_dir)) for f in manifest_files]
        assert 'app1/manifest.yml' in rel_paths

    def test_manifest_exclude_regexes_empty(self, tmp_path):
        """Test with empty exclude patterns"""
        root_dir = tmp_path / 'project'
        root_dir.mkdir()

        # Create manifest files including one in managed_components
        (root_dir / 'app1' / 'manifest.yml').parent.mkdir()
        (root_dir / 'app1' / 'manifest.yml').write_text('app1:\n  enable:\n    - if: IDF_TARGET == "esp32"')

        (root_dir / 'managed_components' / 'component1' / 'manifest.yml').parent.mkdir(parents=True)
        (root_dir / 'managed_components' / 'component1' / 'manifest.yml').write_text(
            'component1:\n  enable:\n    - if: IDF_TARGET == "esp32"'
        )

        os.chdir(root_dir)

        # Test with empty exclude patterns - should include all files
        args = DependencyDrivenBuildArguments(
            manifest_filepatterns=['**/*.yml'], manifest_exclude_regexes=[], manifest_rootpath=str(root_dir)
        )

        manifest_files = args.manifest_files
        assert manifest_files is not None
        assert len(manifest_files) == 2

        rel_paths = [os.path.relpath(f, str(root_dir)) for f in manifest_files]
        assert 'app1/manifest.yml' in rel_paths
        assert 'managed_components/component1/manifest.yml' in rel_paths

    def test_manifest_exclude_regexes_with_existing_manifest_files(self, tmp_path):
        """Test that exclude patterns work when manifest_files are already provided"""
        root_dir = tmp_path / 'project'
        root_dir.mkdir()

        # Create manifest files
        existing_manifest = root_dir / 'existing.yml'
        existing_manifest.write_text('existing:\n  enable:\n    - if: IDF_TARGET == "esp32"')

        (root_dir / 'app1' / 'manifest.yml').parent.mkdir()
        (root_dir / 'app1' / 'manifest.yml').write_text('app1:\n  enable:\n    - if: IDF_TARGET == "esp32"')

        (root_dir / 'managed_components' / 'component1' / 'manifest.yml').parent.mkdir(parents=True)
        (root_dir / 'managed_components' / 'component1' / 'manifest.yml').write_text(
            'component1:\n  enable:\n    - if: IDF_TARGET == "esp32"'
        )

        os.chdir(root_dir)

        # Test with existing manifest files + filepatterns + exclude patterns
        args = DependencyDrivenBuildArguments(
            manifest_files=[str(existing_manifest)],
            manifest_filepatterns=['**/*.yml'],
            manifest_exclude_regexes=['/managed_components/'],
            manifest_rootpath=str(root_dir),
        )

        manifest_files = args.manifest_files
        assert manifest_files is not None
        assert len(manifest_files) == 3  # existing (twice) + app1 (managed_components excluded)

        rel_paths = [os.path.relpath(f, str(root_dir)) for f in manifest_files]
        assert 'existing.yml' in rel_paths
        assert 'app1/manifest.yml' in rel_paths
        assert 'managed_components/component1/manifest.yml' not in rel_paths

    def test_manifest_exclude_regexes_complex_patterns(self, tmp_path):
        """Test complex exclude patterns with wildcards"""
        root_dir = tmp_path / 'project'
        root_dir.mkdir()

        # Create various manifest files
        (root_dir / 'app1' / 'manifest.yml').parent.mkdir()
        (root_dir / 'app1' / 'manifest.yml').write_text('app1:\n  enable:\n    - if: IDF_TARGET == "esp32"')

        (root_dir / 'app2' / 'foo_manifest.yml').parent.mkdir()
        (root_dir / 'app2' / 'foo_manifest.yml').write_text('app2:\n  enable:\n    - if: IDF_TARGET == "esp32"')

        (root_dir / 'docs' / 'example.yml').parent.mkdir()
        (root_dir / 'docs' / 'example.yml').write_text('docs:\n  enable:\n    - if: IDF_TARGET == "esp32"')

        os.chdir(root_dir)

        # Test with complex exclude patterns
        args = DependencyDrivenBuildArguments(
            manifest_filepatterns=['**/*.yml'],
            manifest_exclude_regexes=['/foo_.+', '/docs/'],
            manifest_rootpath=str(root_dir),
        )

        manifest_files = args.manifest_files
        assert manifest_files is not None
        assert len(manifest_files) == 1

        rel_paths = [os.path.relpath(f, str(root_dir)) for f in manifest_files]
        assert 'app1/manifest.yml' in rel_paths
        assert 'app2/foo_manifest.yml' not in rel_paths
        assert 'docs/example.yml' not in rel_paths

    def test_manifest_exclude_regexes_no_filepatterns(self, tmp_path):
        """Test that exclude patterns have no effect when no filepatterns are specified"""
        root_dir = tmp_path / 'project'
        root_dir.mkdir()

        manifest_file = root_dir / 'manifest.yml'
        manifest_file.write_text('app:\n  enable:\n    - if: IDF_TARGET == "esp32"')

        os.chdir(root_dir)

        # Test with only exclude patterns (no filepatterns)
        args = DependencyDrivenBuildArguments(
            manifest_files=[str(manifest_file)], manifest_exclude_regexes=['.+'], manifest_rootpath=str(root_dir)
        )

        # Should still have the explicitly provided manifest file
        manifest_files = args.manifest_files
        assert manifest_files is not None
        assert len(manifest_files) == 1
        assert str(manifest_file) in manifest_files



# SPDX-FileCopyrightText: 2022-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import os.path
import re
import typing as t
from pathlib import (
    Path,
)

from .app import (
    App,
    CMakeApp,
)
from .args import FindArguments
from .constants import (
    ALL_TARGETS,
    BuildStatus,
)
from .utils import (
    config_rules_from_str,
    to_absolute_path,
)

LOGGER = logging.getLogger(__name__)


def _is_target_specific(filepath: str) -> bool:
    for target in ALL_TARGETS:
        if filepath.endswith(f'.{target}'):
            return True

    return False


def _get_apps_from_path(
    path: str,
    target: str,
    *,
    app_cls: t.Type[App] = CMakeApp,
    args: FindArguments,
) -> t.List[App]:
    if not app_cls.is_app(path):
        LOGGER.debug('Skipping. %s is not an app', path)
        return []

    config_rules = config_rules_from_str(args.config_rules)

    apps = []
    app_configs: t.List[t.Tuple[t.Optional[str], str]] = []  # List of (sdkconfig_path, config_name) tuples
    default_config_name = ''
    sdkconfig_paths_matched = False

    for rule in config_rules:
        if not rule.file_name:
            default_config_name = rule.config_name
            continue

        sdkconfig_paths = sorted([str(p.resolve()) for p in Path(path).glob(rule.file_name)])

        if sdkconfig_paths:
            sdkconfig_paths_matched = True  # skip the next block for no wildcard config rules

        for sdkconfig_path in sdkconfig_paths:
            if _is_target_specific(sdkconfig_path):
                LOGGER.debug('=> Skipping sdkconfig file `%s` which is target-specific', sdkconfig_path)
                continue

            # Figure out the config name
            config_name = rule.config_name or ''
            if '*' in rule.file_name:
                # convert glob pattern into a regex
                regex_str = r'.*' + rule.file_name.replace('.', r'\.').replace('*', r'(.*)')
                groups = re.match(regex_str, sdkconfig_path)
                assert groups
                config_name = groups.group(1)

            app_configs.append((sdkconfig_path, config_name))

    # no config rules matched, use default app
    if not sdkconfig_paths_matched:
        app_configs.append((None, default_config_name))

    # Create, validate, and add all apps
    for p, n in app_configs:
        app = app_cls(
            path,
            target,
            sdkconfig_path=p,
            config_name=n,
            work_dir=args.work_dir,
            build_dir=args.build_dir,
            build_log_filename=args.build_log_filename,
            size_json_filename=args.size_json_filename,
            size_json_extra_args=args.size_json_extra_args,
            check_warnings=args.check_warnings,
            sdkconfig_defaults_str=args.sdkconfig_defaults,
        )

        if app.sdkconfig_files_defined_idf_target and app.target != app.sdkconfig_files_defined_idf_target:
            LOGGER.debug(
                'Project %s with config %s defined CONFIG_IDF_TARGET=%s, Ignoring target %s',
                app.app_dir,
                app.config_name or 'default',
                app.sdkconfig_files_defined_idf_target,
                target,
            )
            continue

        app.check_should_build(
            manifest_rootpath=args.manifest_rootpath,
            modified_manifest_rules_folders=args.modified_manifest_rules_folders,
            modified_components=args.modified_components,
            modified_files=args.modified_files,
            check_app_dependencies=args.dependency_driven_build_enabled,
        )
        app.check_should_test()

        if app.build_status == BuildStatus.DISABLED:
            LOGGER.debug('=> Disabled. Reason: %s', app.build_comment or 'Unknown')
            should_include = args.include_disabled_apps
        elif app.build_status == BuildStatus.SKIPPED:
            LOGGER.debug('=> Skipped. Reason: %s', app.build_comment or 'Unknown')
            should_include = args.include_skipped_apps
        else:
            should_include = True

        if should_include:
            LOGGER.debug('Found app: %s', app)
            apps.append(app)

    return sorted(apps)


def _find_apps(
    path: str,
    target: str,
    *,
    app_cls: t.Type[App] = CMakeApp,
    args: FindArguments,
) -> t.List[App]:
    LOGGER.debug(
        'Looking for %s apps in %s%s with target %s',
        app_cls.__name__,
        path,
        ' recursively' if args.recursive else '',
        target,
    )

    if not args.recursive:
        if args.exclude:
            LOGGER.debug('--exclude option is ignored when used without --recursive')

        return _get_apps_from_path(path, target, app_cls=app_cls, args=args)

    # The remaining part is for recursive == True
    apps = []
    # handle the exclude list, since the config file might use linux style, but run in windows
    exclude_paths_list = [to_absolute_path(p) for p in args.exclude or []]
    for root, dirs, _ in os.walk(path):
        LOGGER.debug('Entering %s', root)
        root_path = to_absolute_path(root)
        if root_path in exclude_paths_list:
            LOGGER.debug('=> Skipping %s (excluded)', root)
            del dirs[:]
            continue

        if os.path.basename(root_path) == 'managed_components':  # idf-component-manager
            LOGGER.debug('=> Skipping %s (managed components)', root_path)
            del dirs[:]
            continue

        _found_apps = _get_apps_from_path(root, target, app_cls=app_cls, args=args)
        if _found_apps:  # root has at least one app
            LOGGER.debug('=> Stop iteration sub dirs of %s since it has apps', root)
            del dirs[:]
            apps.extend(_found_apps)
            continue

    return apps



# SPDX-FileCopyrightText: 2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0


from conftest import create_project

from idf_build_apps.constants import BuildStatus
from idf_build_apps.main import find_apps


class TestBuildDisableReason:
    """Test build disable reason functionality with real projects"""

    def test_manifest_disable_rule_sets_reason(self, tmp_path):
        create_project('test_app', tmp_path)

        manifest_file = tmp_path / 'manifest.yml'
        manifest_file.write_text("""
test_app:
    disable:
        - if: IDF_TARGET == "esp32s3"
          reason: "Not supported on this target"
""")

        apps = find_apps(
            str(tmp_path / 'test_app'), 'esp32s3', manifest_files=[str(manifest_file)], include_disabled_apps=True
        )

        assert len(apps) == 1
        app = apps[0]
        assert app.build_status == BuildStatus.DISABLED
        assert (
            'Disabled by manifest rule: IDF_TARGET == "esp32s3" (reason: Not supported on this target)'
            == app.build_comment
        )

    def test_manifest_enable_only_specific_targets(self, tmp_path):
        """Test that apps not in enable list get proper disable reason"""
        create_project('test_app', tmp_path)

        manifest_file = tmp_path / 'manifest.yml'
        manifest_file.write_text("""
test_app:
    enable:
        - if: IDF_TARGET == "esp32"
        - if: IDF_TARGET == "esp32s2"
""")

        apps = find_apps(
            str(tmp_path / 'test_app'), 'esp32c3', manifest_files=[str(manifest_file)], include_disabled_apps=True
        )

        assert len(apps) == 1
        app = apps[0]
        assert app.build_status == BuildStatus.DISABLED
        assert (
            'Not enabled by manifest rules:\n- IDF_TARGET == "esp32"\n- IDF_TARGET == "esp32s2"'
        ) == app.build_comment

    def test_sdkconfig_target_mismatch_excluded(self, tmp_path):
        """Test that CONFIG_IDF_TARGET mismatch sets proper disable reason"""
        create_project('test_app', tmp_path)

        sdkconfig_file = tmp_path / 'test_app' / 'sdkconfig.defaults'
        sdkconfig_file.write_text('CONFIG_IDF_TARGET="esp32"\n')

        apps = find_apps(str(tmp_path / 'test_app'), 'esp32s2', include_disabled_apps=True)

        assert not apps  # this shall not be disabled. shall be excluded.

    def test_target_not_in_default_build_targets(self, tmp_path):
        """Test that targets not in default build targets get proper disable reason"""
        create_project('test_app', tmp_path)

        apps = find_apps(
            str(tmp_path / 'test_app'),
            'unsupported_target',
            default_build_targets=['esp32', 'esp32s2'],
            include_disabled_apps=True,
        )

        assert len(apps) == 1
        app = apps[0]
        assert app.build_status == BuildStatus.DISABLED
        assert 'Target unsupported_target not in default build targets esp32,esp32s2' == app.build_comment


class TestTestDisableReason:
    """Test test_disable_reason functionality with real projects"""

    def test_no_manifest_no_test_disable(self, tmp_path):
        """Test that apps without manifest don't have test disabling"""
        create_project('test_app', tmp_path)

        apps = find_apps(str(tmp_path / 'test_app'), 'esp32')

        assert len(apps) == 1
        app = apps[0]

        app.check_should_test()

        # Without a manifest, testing should not be disabled
        assert app.test_comment is None

    def test_manifest_disable_test_rule_sets_reason(self, tmp_path):
        """Test that manifest disable_test rules set proper test disable reason"""
        create_project('test_app', tmp_path)

        manifest_file = tmp_path / 'manifest.yml'
        manifest_file.write_text("""
test_app:
    disable_test:
        - if: IDF_TARGET == "esp32c3"
          reason: "Testing not stable on this target"
""")

        # Find the app - should be enabled for building but test disabled
        apps = find_apps(str(tmp_path / 'test_app'), 'esp32c3', manifest_files=[str(manifest_file)])

        assert len(apps) == 1
        app = apps[0]

        # App should be buildable
        assert app.build_status != BuildStatus.DISABLED

        # But testing should be disabled
        app.check_should_test()
        assert (
            'Disabled by manifest rule: IDF_TARGET == "esp32c3" (reason: Testing not stable on this target)'
            == app.test_comment
        )

    def test_manifest_general_disable_affects_test(self, tmp_path):
        """Test that general manifest disable rules also affect testing"""
        create_project('test_app', tmp_path)

        manifest_file = tmp_path / 'manifest.yml'
        manifest_file.write_text("""
test_app:
    disable:
        - if: IDF_TARGET == "esp32h2"
          reason: "Target not supported"
""")

        # Find the app with include_disabled_apps=True to get the disabled app
        apps = find_apps(
            str(tmp_path / 'test_app'), 'esp32h2', manifest_files=[str(manifest_file)], include_disabled_apps=True
        )

        assert len(apps) == 1
        app = apps[0]

        # App should be disabled for building
        assert app.build_status == BuildStatus.DISABLED

        # Testing should also be disabled by the same rule
        app.check_should_test()
        assert app.test_comment is not None
        assert 'Disabled by manifest rule: IDF_TARGET == "esp32h2"' in app.test_comment
        assert '(reason: Target not supported)' in app.test_comment

    def test_manifest_enable_specific_targets_for_test(self, tmp_path):
        """Test that targets not enabled by manifest get proper test disable reason"""
        create_project('test_app', tmp_path)

        manifest_file = tmp_path / 'manifest.yml'
        manifest_file.write_text("""
test_app:
    enable:
        - if: IDF_TARGET == "esp32"
        - if: IDF_TARGET == "esp32s2"
""")

        # Find the app for a target not in the enable list
        apps = find_apps(
            str(tmp_path / 'test_app'), 'esp32s3', manifest_files=[str(manifest_file)], include_disabled_apps=True
        )

        assert len(apps) == 1
        app = apps[0]

        # App should be disabled for building
        assert app.build_status == BuildStatus.DISABLED

        # Testing should also be disabled
        app.check_should_test()
        assert (
            'Not enabled by manifest rules:\n- IDF_TARGET == "esp32"\n- IDF_TARGET == "esp32s2"'
        ) == app.test_comment

    def test_mixed_build_and_test_disable_rules(self, tmp_path):
        """Test apps with both build disable and test-specific disable rules"""
        create_project('build_ok_test_disabled', tmp_path)
        create_project('both_disabled', tmp_path)

        manifest_file = tmp_path / 'manifest.yml'
        manifest_file.write_text("""
build_ok_test_disabled:
    disable_test:
        - if: IDF_TARGET == "esp32c3"
          temporary: true
          reason: "Flaky tests on this hardware"

both_disabled:
    disable:
        - if: IDF_TARGET == "esp32c3"
          reason: "No hardware support"
""")

        apps = find_apps(
            str(tmp_path), 'esp32c3', recursive=True, manifest_files=[str(manifest_file)], include_disabled_apps=True
        )

        # Should find 2 apps
        assert len(apps) == 2

        # Sort by app name for consistent testing
        apps.sort(key=lambda x: x.name)

        # both_disabled: should be disabled for building
        both_disabled_app = apps[0]  # both_disabled comes first alphabetically
        assert both_disabled_app.name == 'both_disabled'
        assert both_disabled_app.build_status == BuildStatus.DISABLED
        assert (
            'Disabled by manifest rule: IDF_TARGET == "esp32c3" (reason: No hardware support)'
            == both_disabled_app.build_comment
        )

        # build_ok_test_disabled: should be OK for building but disabled for testing
        build_ok_app = apps[1]  # build_ok_test_disabled comes second alphabetically
        assert build_ok_app.name == 'build_ok_test_disabled'
        assert build_ok_app.build_status != BuildStatus.DISABLED

        build_ok_app.check_should_test()
        assert (
            'Disabled by manifest rule: IDF_TARGET == "esp32c3" (temporary) (reason: Flaky tests on this hardware)'
            == build_ok_app.test_comment
        )



# SPDX-FileCopyrightText: 2022-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

import json
import os
import shutil
import subprocess
from copy import (
    deepcopy,
)
from pathlib import Path
from xml.etree import (
    ElementTree,
)

import pytest
from conftest import (
    create_project,
)

from idf_build_apps import (
    build_apps,
    find_apps,
)
from idf_build_apps.app import (
    App,
    CMakeApp,
)
from idf_build_apps.args import BuildArguments
from idf_build_apps.constants import (
    IDF_PATH,
    BuildStatus,
)
from idf_build_apps.utils import Literal


@pytest.mark.skipif(not shutil.which('idf.py'), reason='idf.py not found')
class TestBuild:
    def test_build_hello_world(self, tmp_path, capsys):
        path = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')

        app = CMakeApp(path, 'esp32', work_dir=str(tmp_path / 'test'))
        app.build()

        captured = capsys.readouterr()
        assert 'Configuring done' in captured.out
        assert 'Project build complete.' in captured.out
        assert app.build_status == BuildStatus.SUCCESS

    @pytest.mark.parametrize(
        'modified_components, check_app_dependencies, build_status',
        [
            (None, True, BuildStatus.SUCCESS),
            ([], True, BuildStatus.SKIPPED),
            ([], False, BuildStatus.SUCCESS),
            ('fake', True, BuildStatus.SKIPPED),
            ('fake', False, BuildStatus.SUCCESS),
            ('soc', True, BuildStatus.SUCCESS),
            ('soc', False, BuildStatus.SUCCESS),
            (['soc', 'fake'], True, BuildStatus.SUCCESS),
        ],
    )
    def test_build_with_modified_components(self, tmp_path, modified_components, check_app_dependencies, build_status):
        path = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')

        app = CMakeApp(path, 'esp32', work_dir=str(tmp_path / 'test'))
        app.build(
            modified_components=modified_components,
            check_app_dependencies=check_app_dependencies,
        )
        assert app.build_status == build_status

    @pytest.mark.parametrize(
        'modified_files, build_status',
        [
            ('/foo', BuildStatus.SKIPPED),
            (os.path.join(IDF_PATH, 'examples', 'README.md'), BuildStatus.SKIPPED),
            ([os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'README.md')], BuildStatus.SKIPPED),
            (
                [
                    os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'README.md'),
                    os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world', 'main', 'hello_world_main.c'),
                ],
                BuildStatus.SUCCESS,
            ),
        ],
    )
    def test_build_with_modified_files(self, modified_files, build_status):
        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')

        app = CMakeApp(test_dir, 'esp32')
        app.build(
            modified_components=[],
            modified_files=modified_files,
            check_app_dependencies=True,
        )

        assert app.build_status == build_status

    def test_build_without_modified_components_but_ignored_app_dependency_check(self):
        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')

        apps = find_apps(
            test_dir,
            'esp32',
            modified_components=[],
            modified_files=['foo.c'],
            ignore_app_dependencies_filepatterns=['foo.c'],
        )

        for app in apps:
            app.build()
            assert app.build_status == BuildStatus.SUCCESS

    def test_build_with_junit_output(self, tmp_path):
        test_dir = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')

        apps = [
            CMakeApp(test_dir, 'esp32', build_dir='build_1'),
            CMakeApp(test_dir, 'esp32', build_dir='build_2'),
            CMakeApp(test_dir, 'esp32', build_dir='build_3'),
            CMakeApp(test_dir, 'esp32', build_dir='build_4'),
        ]
        apps[2].build_status = BuildStatus.DISABLED
        apps[3].build_status = BuildStatus.SKIPPED

        build_apps(deepcopy(apps), dry_run=True, junitxml=str(tmp_path / 'test.xml'))

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())

        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['tests'] == '0'
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '4'

        for i, testcase in enumerate(test_suite.findall('testcase')):
            assert testcase.attrib['name'] == apps[i].build_path
            assert float(testcase.attrib['time']) > 0
            assert testcase.find('skipped') is not None
            if i in (0, 1):
                assert testcase.find('skipped').attrib['message'] == 'dry run'
            elif i == 2:
                assert testcase.find('skipped').attrib['message'] == 'Build disabled. Skipping...'
            elif i == 3:
                assert testcase.find('skipped').attrib['message'] == 'Build skipped. Skipping...'
            else:
                assert False  # not expected

        build_apps(deepcopy(apps), junitxml=str(tmp_path / 'test.xml'))

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())

        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['tests'] == '2'
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '2'

        for i, testcase in enumerate(test_suite.findall('testcase')):
            assert float(testcase.attrib['time']) > 0
            assert testcase.attrib['name'] == apps[i].build_path
            assert testcase.find('error') is None
            assert testcase.find('failure') is None
            if i in (0, 1):
                assert testcase.find('skipped') is None
            elif i == 2:
                assert testcase.find('skipped').attrib['message'] == 'Build disabled. Skipping...'
            elif i == 3:
                assert testcase.find('skipped').attrib['message'] == 'Build skipped. Skipping...'
            else:
                assert False  # not expected

    def test_work_dir_inside_relative_app_dir(self, tmp_path):
        create_project('foo', tmp_path)

        os.chdir(tmp_path / 'foo')
        apps = find_apps(
            '.',
            'esp32',
            work_dir=os.path.join('foo', 'bar'),
        )
        build_apps(apps)

        assert len(apps) == 1
        assert apps[0].build_status == BuildStatus.SUCCESS

    def test_build_apps_without_passing_apps(self, tmp_path):
        create_project('foo', tmp_path)

        os.chdir(tmp_path / 'foo')
        ret_code = build_apps(
            target='esp32',
            work_dir=os.path.join('foo', 'bar'),
            junitxml='test.xml',
        )
        assert ret_code == 0

        with open('test.xml') as f:
            xml = ElementTree.fromstring(f.read())

        test_suite = xml.findall('testsuite')[0]
        assert test_suite.attrib['tests'] == '1'
        assert test_suite.attrib['failures'] == '0'
        assert test_suite.attrib['errors'] == '0'
        assert test_suite.attrib['skipped'] == '0'

        assert test_suite.findall('testcase')[0].attrib['name'] == 'foo/bar/build'

    def test_build_with_size_json_without_extra_args(self, tmp_path):
        """Test idf_size_py without extra args."""
        path = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')

        app = CMakeApp(path, 'esp32', work_dir=str(tmp_path / 'test'), size_json_filename='size.json')
        app.build()
        app.write_size_json()

        assert app.build_status == BuildStatus.SUCCESS
        assert os.path.exists(app.size_json_path)

        with open(app.size_json_path) as f:
            size_data = json.load(f)

        expected_keys = ['dram_data', 'dram_bss', 'used_dram', 'iram_text', 'flash_code', 'total_size']
        for key in expected_keys:
            assert key in size_data, f"Size JSON should contain '{key}' when no extra args"

    def test_build_with_size_json_with_extra_args(self, tmp_path):
        """Test idf_size_py with extra args."""
        path = os.path.join(IDF_PATH, 'examples', 'get-started', 'hello_world')

        app = CMakeApp(
            path,
            'esp32',
            work_dir=str(tmp_path / 'test'),
            size_json_filename='size.json',
            size_json_extra_args=['--files'],
        )
        app.build()
        app.write_size_json()

        assert app.build_status == BuildStatus.SUCCESS
        assert os.path.exists(app.size_json_path)

        with open(app.size_json_path) as f:
            size_data = json.load(f)

        expected_keys = [
            'libesp_timer.a:system_time.c.obj',
            'libc.a:libc_a-vfprintf.o',
            'libfreertos.a:tasks.c.obj',
            'libheap.a:tlsf.c.obj',
        ]
        for key in expected_keys:
            assert key in size_data, f"Size JSON should contain '{key}' with extra args"


class CustomClassApp(App):
    build_system: Literal['custom_class'] = 'custom_class'  # type: ignore

    def build(self, *args, **kwargs):
        # For testing, we'll just create a dummy build directory
        if not self.dry_run:
            os.makedirs(self.build_path, exist_ok=True)
            with open(os.path.join(self.build_path, 'dummy.txt'), 'w') as f:
                f.write('Custom build successful')
        self.build_status = BuildStatus.SUCCESS
        print('Custom build successful')

    @classmethod
    def is_app(cls, path: str) -> bool:  # noqa: ARG003
        return True


@pytest.mark.skipif(not shutil.which('idf.py'), reason='idf.py not found')
class TestBuildWithCustomApp:
    custom_app_code = """
from idf_build_apps import App
import os
from idf_build_apps.constants import BuildStatus
from idf_build_apps.utils import Literal

class CustomApp(App):
    build_system: Literal['custom'] = 'custom'

    def build(self, *args, **kwargs):
        # For testing, we'll just create a dummy build directory
        if not self.dry_run:
            os.makedirs(self.build_path, exist_ok=True)
            with open(os.path.join(self.build_path, 'dummy.txt'), 'w') as f:
                f.write('Custom build successful')
        self.build_status = BuildStatus.SUCCESS
        print('Custom build successful')

    @classmethod
    def is_app(cls, path: str) -> bool:
        return True
    """

    @pytest.fixture(autouse=True)
    def _setup(self, tmp_path: Path, monkeypatch):
        os.chdir(tmp_path)

        test_app = tmp_path / 'test_app'

        test_app.mkdir()
        (test_app / 'main' / 'main.c').parent.mkdir(parents=True)
        (test_app / 'main' / 'main.c').write_text('void app_main() {}')

        # Create a custom app module
        custom_module = tmp_path / 'custom.py'
        custom_module.write_text(self.custom_app_code)

        monkeypatch.setenv('PYTHONPATH', os.getenv('PYTHONPATH', '') + os.pathsep + str(tmp_path))

        return test_app

    def test_custom_app_cli(self, tmp_path):
        subprocess.run(
            [
                'idf-build-apps',
                'build',
                '-p',
                'test_app',
                '--target',
                'esp32',
                '--build-system',
                'custom:CustomApp',
            ],
            check=True,
        )

        assert (tmp_path / 'test_app' / 'build' / 'dummy.txt').exists()
        assert (tmp_path / 'test_app' / 'build' / 'dummy.txt').read_text() == 'Custom build successful'

    def test_custom_app_function(self, tmp_path):
        # Import the custom app class
        # Find and build the app using the imported CustomApp class
        apps = find_apps(
            paths=['test_app'],
            target='esp32',
            build_system=CustomClassApp,
        )

        assert len(apps) == 1
        app = apps[0]
        assert isinstance(app, CustomClassApp)
        assert app.build_system == 'custom_class'

        # Build the app
        app.build()
        assert app.build_status == BuildStatus.SUCCESS
        assert (tmp_path / 'test_app' / 'build' / 'dummy.txt').exists()
        assert (tmp_path / 'test_app' / 'build' / 'dummy.txt').read_text() == 'Custom build successful'


def test_build_apps_collect_files_when_no_apps_built(tmp_path):
    os.chdir(tmp_path)

    build_apps(
        build_arguments=BuildArguments(
            target='esp32',
            collect_app_info_filename='app_info.txt',
            collect_size_info_filename='size_info.txt',
        )
    )

    assert os.path.exists('app_info.txt')
    assert os.path.exists('size_info.txt')



# SPDX-FileCopyrightText: 2022-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0
import contextvars
import logging
import os
import typing as t
import warnings
from functools import lru_cache
from hashlib import sha512

from esp_bool_parser import BoolStmt, parse_bool_expr
from pyparsing import (
    ParseException,
)

from ..constants import (
    ALL_TARGETS,
    SUPPORTED_TARGETS,
)
from ..utils import (
    InvalidIfClause,
    InvalidManifest,
    PathLike,
    to_absolute_path,
)
from ..yaml import (
    parse,
)

LOGGER = logging.getLogger(__name__)

# Context variable for default build targets
DEFAULT_BUILD_TARGETS: contextvars.ContextVar[t.List[str]] = contextvars.ContextVar(
    'default_build_targets', default=SUPPORTED_TARGETS
)


def reset_default_build_targets() -> None:
    """Reset DEFAULT_BUILD_TARGETS to the default value (SUPPORTED_TARGETS)"""
    DEFAULT_BUILD_TARGETS.set(SUPPORTED_TARGETS)


class IfClause:
    def __init__(self, stmt: str, temporary: bool = False, reason: t.Optional[str] = None) -> None:
        try:
            self.stmt: BoolStmt = parse_bool_expr(stmt)
            self._stmt: str = stmt
        except (ParseException, InvalidIfClause) as ex:
            raise InvalidIfClause(f'Invalid if clause: {stmt}. {ex}')

        self.temporary = temporary
        self.reason = reason

        if self.temporary is True and not self.reason:
            raise InvalidIfClause(
                f'Invalid if clause "{stmt}". '
                f'"reason" must be set when "temporary: true". '
                f'For example:\n'
                f'  - if: {stmt}\n'
                f'    temporary: true\n'
                f'    reason: lack of ci runners'
            )

    def __str__(self):
        s = self._stmt
        if self.temporary:
            s += ' (temporary)'
        if self.reason:
            s += f' (reason: {self.reason})'
        return s

    def __repr__(self):
        return f'IfClause(stmt={self._stmt!r}, temporary={self.temporary!r}, reason={self.reason!r})'

    def get_value(self, target: str, config_name: str) -> t.Any:
        return self.stmt.get_value(target, config_name)


class SwitchClause:
    def __init__(
        self, if_clauses: t.List[IfClause], contents: t.List[t.List[str]], default_clause: t.List[str]
    ) -> None:
        self.if_clauses = if_clauses
        self.contents = contents
        self.default_clause = default_clause

    def get_value(self, target: str, config_name: str) -> t.Any:
        for if_clause, content in zip(self.if_clauses, self.contents):
            if if_clause.get_value(target, config_name):
                return content
        return self.default_clause

    def __repr__(self) -> str:
        return (
            f'SwitchClause('
            f'if_clauses={self.if_clauses!r}, '
            f'contents={self.contents!r}, '
            f'default_clause={self.default_clause!r})'
        )


def _getattr_default_build_targets(name: str) -> t.Any:
    if name == 'DEFAULT_BUILD_TARGETS':
        warnings.warn(
            'FolderRule.DEFAULT_BUILD_TARGETS is deprecated. Use DEFAULT_BUILD_TARGETS.get() directly.',
            DeprecationWarning,
            stacklevel=3,
        )
        return DEFAULT_BUILD_TARGETS.get()
    return None


def _setattr_default_build_targets(name: str, value: t.Any) -> bool:
    if name == 'DEFAULT_BUILD_TARGETS':
        warnings.warn(
            'FolderRule.DEFAULT_BUILD_TARGETS is deprecated. Use DEFAULT_BUILD_TARGETS.set() directly.',
            DeprecationWarning,
            stacklevel=3,
        )
        if not isinstance(value, list):
            raise TypeError('Default build targets must be a list')
        DEFAULT_BUILD_TARGETS.set(value)
        return True
    return False


class _FolderRuleMeta(type):
    """Metaclass to handle class-level assignments to DEFAULT_BUILD_TARGETS"""

    def __getattribute__(cls, name):
        result = _getattr_default_build_targets(name)
        if result is not None:
            return result
        return super().__getattribute__(name)

    def __setattr__(cls, name, value):
        if _setattr_default_build_targets(name, value):
            return
        super().__setattr__(name, value)

    def __delattr__(cls, name):
        if name == 'DEFAULT_BUILD_TARGETS':
            # Don't actually delete anything, just ignore the deletion
            # This handles monkeypatch teardown issues
            pass
        else:
            super().__delattr__(name)


class FolderRule(metaclass=_FolderRuleMeta):
    def __getattribute__(self, name):  # instance attr
        result = _getattr_default_build_targets(name)
        if result is not None:
            return result
        return super().__getattribute__(name)

    def __setattr__(self, name, value):
        if _setattr_default_build_targets(name, value):
            return
        super().__setattr__(name, value)

    def __init__(
        self,
        folder: str,
        enable: t.Optional[t.List[t.Dict[str, t.Any]]] = None,
        disable: t.Optional[t.List[t.Dict[str, t.Any]]] = None,
        disable_test: t.Optional[t.List[t.Dict[str, t.Any]]] = None,
        depends_components: t.Optional[t.List[t.Union[str, t.Dict[str, t.Any]]]] = None,
        depends_filepatterns: t.Optional[t.List[t.Union[str, t.Dict[str, t.Any]]]] = None,
        manifest_filepath: t.Optional[str] = None,
    ) -> None:
        self._manifest_filepath = manifest_filepath

        self.folder = os.path.abspath(folder)

        def _clause_to_if_clause(clause: t.Dict[str, t.Any]) -> IfClause:
            _kwargs = {'stmt': clause['if']}
            if 'temporary' in clause:
                _kwargs['temporary'] = clause['temporary']
            if 'reason' in clause:
                _kwargs['reason'] = clause['reason']
            return IfClause(**_kwargs)

        def _clause_to_switch_or_list(
            statements: t.Optional[t.List[t.Union[str, t.Dict[str, t.Any]]]],
        ) -> t.Union[SwitchClause, t.List[str]]:
            if not statements:
                return []

            switch_statements = []
            str_statements = []
            for statement in statements:
                if isinstance(statement, t.Dict):
                    switch_statements.append(statement)
                else:
                    str_statements.append(statement)

            if switch_statements and str_statements:
                raise InvalidManifest('Current manifest format has to fit either the switch format or the list format.')

            if str_statements:
                return str_statements

            return _clause_to_switch_clause(switch_statements)

        def _clause_to_switch_clause(switch_statements: t.List[t.Dict[str, t.Any]]) -> SwitchClause:
            if_clauses = []
            contents = []
            default_clauses = []
            for statement in switch_statements:
                if 'if' in statement:
                    if_clauses.append(IfClause(stmt=statement['if']))
                    contents.append(statement['content'])
                elif 'default' in statement:
                    default_clauses.extend(statement['default'])
                else:
                    raise InvalidManifest("Only the 'if' and 'default' keywords are supported in switch clause.")

            return SwitchClause(if_clauses, contents, default_clauses)

        self.enable = [_clause_to_if_clause(clause) for clause in enable] if enable else []
        self.disable = [_clause_to_if_clause(clause) for clause in disable] if disable else []
        self.disable_test = [_clause_to_if_clause(clause) for clause in disable_test] if disable_test else []
        self.depends_components = _clause_to_switch_or_list(depends_components)
        self.depends_filepatterns = _clause_to_switch_or_list(depends_filepatterns)

    def __hash__(self) -> int:
        return hash(self.sha)

    @property
    def sha(self) -> str:
        """
        SHA of the FolderRule instance

        :return: SHA value
        """
        sha = sha512()
        for obj in [
            self.enable,
            self.disable,
            self.disable_test,
            self.depends_components,
            self.depends_filepatterns,
        ]:
            sha.update(repr(obj).encode())

        return sha.hexdigest()

    def __repr__(self) -> str:
        return f'FolderRule({self.folder})'

    @property
    def by_manifest_file(self) -> t.Optional[str]:
        return self._manifest_filepath

    @lru_cache(None)
    def _enable_build(self, target: str, config_name: str) -> bool:
        if self.enable:
            res = False
            for clause in self.enable:
                if clause.get_value(target, config_name):
                    res = True
                    break
        else:
            res = target in DEFAULT_BUILD_TARGETS.get()

        if self.disable:
            for clause in self.disable:
                if clause.get_value(target, config_name):
                    res = False
                    break

        return res

    @lru_cache(None)
    def _enable_test(
        self, target: str, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> bool:
        res = target in self.enable_build_targets(default_sdkconfig_target, config_name)

        if self.disable or self.disable_test:
            for clause in self.disable + self.disable_test:
                if clause.get_value(target, config_name or ''):
                    res = False
                    break

        return res

    @lru_cache(None)
    def enable_build_targets(
        self, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> t.List[str]:
        res = []
        for target in ALL_TARGETS:
            if self._enable_build(target, config_name or ''):
                res.append(target)

        if default_sdkconfig_target:
            if default_sdkconfig_target not in res:
                LOGGER.debug(
                    'sdkconfig defined `CONFIG_IDF_TARGET=%s` is not enabled for folder %s. Skip building this App...',
                    default_sdkconfig_target,
                    self.folder,
                )
                return []
            else:
                LOGGER.debug(
                    'sdkconfig defined `CONFIG_IDF_TARGET=%s` overrides the supported targets for folder %s',
                    default_sdkconfig_target,
                    self.folder,
                )
                res = [default_sdkconfig_target]

        return sorted(res)

    @lru_cache(None)
    def enable_test_targets(
        self, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> t.List[str]:
        res = []
        for target in ALL_TARGETS:
            if self._enable_test(target, default_sdkconfig_target, config_name):
                res.append(target)

        return sorted(res)


class DefaultRule(FolderRule):
    def __init__(self, folder: str) -> None:
        super().__init__(folder)


class Manifest:
    # could be reassigned later
    CHECK_MANIFEST_RULES = False

    def __init__(self, rules: t.Iterable[FolderRule], *, root_path: str = os.curdir) -> None:
        self.rules = sorted(rules, key=lambda x: x.folder)
        self._root_path = to_absolute_path(root_path)

        # Pre-compute rule paths
        self._rule_paths = {rule.folder: rule for rule in self.rules}

    @classmethod
    def from_files(cls, paths: t.Iterable[PathLike], *, root_path: str = os.curdir) -> 'Manifest':
        """
        Create a Manifest instance from multiple manifest files

        :param paths: manifest file paths
        :param root_path: root path for relative paths in manifest files
        :return: Manifest instance
        """
        # folder, defined as dict
        _known_folders: t.Dict[str, PathLike] = dict()

        rules: t.List[FolderRule] = []
        for path in paths:
            LOGGER.debug('Loading manifest file %s', path)
            _manifest = cls.from_file(path, root_path=root_path)

            for rule in _manifest.rules:
                if rule.folder in _known_folders:
                    msg = f'Folder "{rule.folder}" is already defined in {_known_folders[rule.folder]}'
                    if cls.CHECK_MANIFEST_RULES:
                        raise InvalidManifest(msg)
                    else:
                        LOGGER.warning(msg)

                _known_folders[rule.folder] = path

            rules.extend(_manifest.rules)

        return Manifest(rules, root_path=root_path)

    @classmethod
    def from_file(cls, path: PathLike, *, root_path: str = os.curdir) -> 'Manifest':
        """
        Create a Manifest instance from a manifest file

        :param path: path to the manifest file
        :param root_path: root path for relative paths in manifest file
        :return: Manifest instance
        """
        manifest_dict = parse(path)

        rules: t.List[FolderRule] = []
        for folder, folder_rule in manifest_dict.items():
            # not a folder, but an anchor
            if folder.startswith('.'):
                continue

            if not os.path.isabs(folder):
                folder = os.path.join(root_path, folder)

            if not os.path.exists(folder):
                msg = f'Folder "{folder}" does not exist. Please check your manifest file {path}'
                if cls.CHECK_MANIFEST_RULES:
                    raise InvalidManifest(msg)
                else:
                    LOGGER.warning(msg)

            try:
                rules.append(FolderRule(folder, **folder_rule if folder_rule else {}, manifest_filepath=str(path)))
            except InvalidIfClause as e:
                raise InvalidManifest(f'Invalid manifest file {path}: {e}')

        return Manifest(rules, root_path=root_path)

    def dump_sha_values(self, sha_filepath: str) -> None:
        """
        Dump the (relative path of the folder, SHA of the FolderRule instance) pairs
        for all rules to the file in format: ``<relative_path>:<SHA>``

        :param sha_filepath: output file path
        :return: None
        """
        with open(sha_filepath, 'w') as fw:
            fw.writelines(f'{os.path.relpath(rule.folder, self._root_path)}:{rule.sha}\n' for rule in self.rules)

    def diff_sha_with_filepath(self, sha_filepath: str, use_abspath: bool = False) -> t.Set[str]:
        """
        Compare the SHA recorded in the file with the current Manifest instance.

        :param sha_filepath: dumped SHA file path
        :param use_abspath: whether to return the absolute path of the folders
        :return: Set of folders that have different SHA values
        """
        recorded__rel_folder__sha__dict = dict()
        with open(sha_filepath) as fr:
            for line in fr:
                line = line.strip()
                if line:
                    try:
                        folder, sha_value = line.strip().rsplit(':', maxsplit=1)
                    except ValueError:
                        raise InvalidManifest(f'Invalid line in SHA file: {line}. Expected format: <folder>:<SHA>')

                    recorded__rel_folder__sha__dict[folder] = sha_value

        self__rel_folder__sha__dict = {os.path.relpath(rule.folder, self._root_path): rule.sha for rule in self.rules}

        diff_folders = set()
        if use_abspath:

            def _path(x):
                return os.path.join(self._root_path, x)
        else:

            def _path(x):
                return x

        for folder, sha_value in recorded__rel_folder__sha__dict.items():
            # removed
            if folder not in self__rel_folder__sha__dict:
                diff_folders.add(_path(folder))
            # modified
            elif sha_value != self__rel_folder__sha__dict[folder]:
                diff_folders.add(_path(folder))

        # new
        for folder in self__rel_folder__sha__dict:
            if folder not in recorded__rel_folder__sha__dict:
                diff_folders.add(_path(folder))

        return diff_folders

    def most_suitable_rule(self, _folder: str) -> FolderRule:
        folder = to_absolute_path(_folder)

        while True:
            if folder in self._rule_paths:
                return self._rule_paths[folder]
            folder = os.path.dirname(folder)

            # reached the root path or the system root path, stop searching
            if folder == self._root_path or os.path.dirname(folder) == folder:
                if folder in self._rule_paths:
                    return self._rule_paths[folder]
                break

        return DefaultRule(folder)

    def enable_build_targets(
        self, folder: str, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> t.List[str]:
        return self.most_suitable_rule(folder).enable_build_targets(default_sdkconfig_target, config_name)

    def enable_test_targets(
        self, folder: str, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> t.List[str]:
        return self.most_suitable_rule(folder).enable_test_targets(default_sdkconfig_target, config_name)

    def depends_components(
        self, folder: str, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> t.List[str]:
        res = self.most_suitable_rule(folder).depends_components
        if isinstance(res, list):
            return res
        return res.get_value(default_sdkconfig_target or '', config_name or '')

    def depends_filepatterns(
        self, folder: str, default_sdkconfig_target: t.Optional[str] = None, config_name: t.Optional[str] = None
    ) -> t.List[str]:
        res = self.most_suitable_rule(folder).depends_filepatterns
        if isinstance(res, list):
            return res
        return res.get_value(default_sdkconfig_target or '', config_name or '')



# SPDX-FileCopyrightText: 2022-2025 Espressif Systems (Shanghai) CO LTD
# SPDX-License-Identifier: Apache-2.0

"""
Manifest file
"""

from .manifest import DEFAULT_BUILD_TARGETS, FolderRule

__all__ = [
    'DEFAULT_BUILD_TARGETS',
    'FolderRule',
]

from esp_bool_parser import register_addition_attribute


def folder_rule_attr(target, **kwargs):
    return 1 if target in DEFAULT_BUILD_TARGETS.get() else 0


register_addition_attribute('INCLUDE_DEFAULT', folder_rule_attr)
